@@ -9,13 +9,15 @@ import scipy.signal
 import scipy.ndimage
 import scipy.cluster.hierarchy
 from scipy.interpolate import UnivariateSpline
+from scipy.ndimage import gaussian_filter
 
 import matplotlib
 import matplotlib.pyplot as plt
 import matplotlib.patheffects as path_effects
 from matplotlib import cm
 
-import plotly.graph_objects
+import plotly.express as px
+import plotly.graph_objects as go
 from plotly.offline import plot as plot_offline
 from plotly.offline import plot_mpl
 from adjustText import adjust_text
@@ -26,13 +28,14 @@ class VisualizationFunctions:
 
     '''Class of visualization functions for DigitalCellSorter'''
 
-    def __init__(self, dataName = 'dataName', saveDir = os.path.join(''), matplotlibMode = 'Agg'):
+    def __init__(self, dataName = 'dataName', saveDir = os.path.join(''), matplotlibMode = 'Agg', safePlotting = True):
 
         '''Function called automatically'''
 
         self.saveDir = saveDir
         self.dataName = dataName
         self.matplotlibMode = matplotlibMode
+        self.safePlotting = safePlotting
 
         return
 
@@ -51,6 +54,22 @@ class VisualizationFunctions:
 
         return
 
+    def tryExcept(func):
+
+      def internal(self, *args, **kwargs):
+
+        if self.safePlotting:
+            try:
+              func(self, *args, **kwargs)
+
+            except Exception as exception:
+              print('Something went wrong while making plot: %s' % (func))
+              print('\tError message: %s\n' % (exception))
+        else:
+            func(self, *args, **kwargs)
+
+      return internal
+
     def saveFigure(self, fig, saveDir, label = 'Figure', extension = 'png', dpi = 300, close = True, attemptSavingHTML = False):
 
         '''Function used internally to save and close figures
@@ -120,6 +139,9 @@ class VisualizationFunctions:
 
         return
 
+    # MatPlotLib-powered figures
+
+    @tryExcept
     def makeHeatmapGeneExpressionPlot(self, df = None, genes = None, nameToAppend = 'heatmap', dpi = 300, extension = 'png', **kwargs):
 
         '''Make heatmap gene expression plot from a provided gene expression matrix.
@@ -215,7 +237,8 @@ class VisualizationFunctions:
 
         return fig
     
-    def makeMarkerExpressionPlot(self, fontscale = 1., dpi = 300, extension = 'png', **kwargs):
+    @tryExcept
+    def makeMarkerExpressionPlot(self, fontscale = 2., dpi = 300, extension = 'png', **kwargs):
 
         '''Produce image on marker genes and their expression on all clusters.
         Uses files generated by function DCS.Vote
@@ -384,6 +407,7 @@ class VisualizationFunctions:
 
         return fig
 
+    @tryExcept
     def internalMakeMarkerSubplots(self, df, X_projection, hugo_cd_dict, NoFrameOnFigures = False, HideClusterLabels = False, outlineClusters = True, analyzeBy = 'cluster', saveSubDir = 'marker_subplots', dpi = 300, extension = 'png', **kwargs):
 
         '''Produce subplots on each marker and its expression on all clusters
@@ -510,6 +534,7 @@ class VisualizationFunctions:
 
         return
 
+    @tryExcept
     def makeAnnotationResultsMatrixPlot(self, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce voting results voting matrix plot
@@ -610,6 +635,7 @@ class VisualizationFunctions:
  
         return fig
 
+    @tryExcept
     def makeHistogramNullDistributionPlot(self, dpi = 600, extension = 'png', **kwargs):
 
         '''Produce histogram plot of the voting null distributions
@@ -740,6 +766,7 @@ class VisualizationFunctions:
 
         return fig
 
+    @tryExcept
     def makeProjectionPlot(self, Xprojection, cellClusterIndexLabel, suffix = '', colormap = cm.jet, legend = True, labels = True, colorbar = False, fontsize = 10, plotNaNs = True, rightShift = 0.3, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce projection plot (2D layout) with a specified coloring scheme
@@ -866,6 +893,7 @@ class VisualizationFunctions:
 
         return fig
 
+    @tryExcept
     def makeStackedBarplot(self, clusterName = None, legendStyle = False, includeLowQC = True, dpi = 300, extension = 'png', **kwargs):
         
         '''Produce stacked barplot with cell fractions
@@ -1006,6 +1034,7 @@ class VisualizationFunctions:
 
         return fig
     
+    @tryExcept
     def makeQualityControlHistogramPlot(self, subset, cutoff, plotPathAndName = None, N_bins = 100, mito = False, displayMeasures = True, precision = 4, quantilePlotCutoff = 0.95, dpi = 300, extension = 'png', fontScale = 1.5, includeTitle = False, **kwargs):
 
         '''Function to calculate QC quality cutoff and visualize it on a histogram
@@ -1162,96 +1191,7 @@ class VisualizationFunctions:
 
         return fig
 
-    def makeSankeyDiagram(self, df, colormapForIndex = None, colormapForColumns = None, linksColor = 'rgba(100,100,100,0.6)', title = '', attemptSavingHTML = False, quality = 4, nameAppend = '_Sankey_diagram'):
-
-        '''Make a Sankey diagram, also known as 'river plot' with two groups of nodes
-
-        Parameters:
-            df: pandas.DataFrame 
-                With counts (overlaps)
-
-            colormapForIndex: dictionary, Default None
-                Colors to use for nodes specified in the DataFrame index
-
-            colormapForColumns: dictionary, Default None
-                Colors to use for nodes specified in the DataFrame columns
-
-            linksColor: str, Default 'rgba(100,100,100,0.6)'
-                Color of the non-overlapping links
-
-            title: str, Default ''
-                Title to print on the diagram
-
-            interactive: boolean , Default False
-                Whether to launch interactive JavaScript-based graph
-
-            quality: int, Default 4
-                Proportional to the resolution of the figure to save
-
-            nameAppend: str, Default '_Sankey_diagram'
-                Name to append to the figure file
-
-        Returns:
-            None
-        
-        Usage:
-            DCS = DigitalCellSorter.DigitalCellSorter()
-
-            DCS.makeSankeyDiagram(df)
-        '''
-
-        try:
-            temp_index = pd.MultiIndex.from_arrays([df.index, [colormapForIndex[item] for item in df.index]], names=['label', 'color'])
-            temp_columns = pd.MultiIndex.from_arrays([df.columns, [colormapForColumns[item] for item in df.columns]], names=['label', 'color'])
-            df.index = temp_index
-            df.columns = temp_columns
-        except Exception as exception:
-            print(exception)
-            print('Using default node colors')
-            colormapForIndex = None
-            colormapForColumns = None
-
-        if (colormapForIndex is None) or (colormapForColumns is None):
-            nodeColors = ['rgba(150,0,10,0.8)'] * len(df.index) + ['rgba(10,0,150,0.8)'] * len(df.columns)
-            nodeLabels = df.index.to_list() + df.columns.to_list()
-        else:
-            nodeLabels = df.index.get_level_values('label').to_list() + df.columns.get_level_values('label').to_list()
-            nodeColors = df.index.get_level_values('color').to_list() + df.columns.get_level_values('color').to_list()
-
-        sources, targets, values, labels = [], [], [], []
-        for i, item in enumerate(df.index):
-            sources.extend([i] * len(df.loc[item]))
-            targets.extend(list(range(len(df.index), len(df.index) + len(df.loc[item]))))
-            values.extend([j for j in df.loc[item].values])
-            #labels.extend([item + ' -> ' + jtem for jtem in df.loc[item].index])
-
-        colorscales = [dict(label=label, colorscale=[[0, linksColor], [1, linksColor]]) for label in labels]
-
-        if not nodeColors is None:
-            for i in range(len(sources)):
-                if nodeColors[sources[i]] == nodeColors[targets[i]]:
-                    newColor = ','.join(nodeColors[sources[i]].split(',')[:3] + ['0.6)'])
-                    colorscales[i] = dict(label=labels[i], colorscale=[[0, newColor], [1, newColor]])
-
-        fig = plotly.graph_objects.Figure(data=[plotly.graph_objects.Sankey(valueformat = '', valuesuffix = '',
-            node = dict(pad = 20, thickness = 40, line = dict(color = 'white', width = 0.5), label = nodeLabels, color = nodeColors,),
-            link = dict(source = sources, target = targets, value = values, label = labels, colorscales = colorscales, hoverinfo='all'))]) #line ={'color':'rgba(255,0,0,0.8)', 'width':0.1}
-
-        if not title is None:
-            fig.update_layout(title_text=title, font_size=10)
-
-        try:
-            fig.write_image(os.path.join(self.saveDir, self.dataName + nameAppend + '.png'), width=500, height=500, scale=quality)
-
-        except Exception as exception:
-            print('Cannot save static image (likely due to missing orca). Saving to interactive html')
-            attemptSavingHTML = True
-
-        if attemptSavingHTML:
-            plot_offline(fig, filename=os.path.join(self.saveDir, self.dataName + nameAppend + '.html'), auto_open=False)
-
-        return fig
-
+    @tryExcept
     def makePlotOfNewMarkers(self, df_marker_cell_type, df_new_marker_cell_type, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce plot of the new markers extracted from the annotated clusters
@@ -1313,6 +1253,7 @@ class VisualizationFunctions:
 
         return fig
 
+    @tryExcept
     def makeTtestPlot(self, df, dfp, label = None, reorder = True, p_value_cutoff = 0.05, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce heatmap plot of t-test p-Values calculated gene-pair-wise
@@ -1413,6 +1354,7 @@ class VisualizationFunctions:
 
         return fig
     
+    @tryExcept
     def makeCellMarkersPiePlot(self, type1, type2, df_marker_cell_type = 'all', nameToAppend = None, listUnexpressedMarkers = True, orthogonalSectorsShift = 0.1, rotationAngle = 0, dpi = 300, extension = 'png', **kwargs):
 
         '''Make summary of markers comparison between two cell types.
@@ -1683,6 +1625,7 @@ class VisualizationFunctions:
 
         return dict(zip(labels, list(sets))), fig
 
+    @tryExcept
     def makeHopfieldPCplot(self, colormap = cm.hot_r, plotTrLines = False, clusterid = 1, trID = 0, axisOff = False, fontscale = 1., trPath = None, dpi = 300, extension = 'png', **kwargs):
 
         '''Make radar plot of the attractors in their principal components coordinates
@@ -1829,7 +1772,8 @@ class VisualizationFunctions:
 
         return fig
 
-    def makeHopfieldLandscapePlot(self, legend = False, labels = False, PCx = 0, PCy = 1, colorbar = True, fontsize = 10, plotMesh = True, plotAttractors = True, adjustText = True, axisOff = True, colorbarva = 0.75, colorbarha = 0.85, trPath = None, colormap = matplotlib.colors.LinearSegmentedColormap.from_list('cmap', [(1, 1, 1), (0, 1, 1), (0, 0, 1), (1, 0, 0)], N=1000), dpi = 300, extension = 'png', **kwargs):
+    @tryExcept
+    def HopfieldLandscapePlot(self, legend = False, labels = False, PCx = 0, PCy = 1, colorbar = True, fontsize = 10, plotMesh = True, plotAttractors = True, adjustText = True, axisOff = True, colorbarva = 0.75, colorbarha = 0.85, trPath = None, colormap = matplotlib.colors.LinearSegmentedColormap.from_list('cmap', [(1, 1, 1), (0, 1, 1), (0, 0, 1), (1, 0, 0)], N=1000), dpi = 300, extension = 'png', **kwargs):
 
         '''Make heatmap plot of the attractors in their two principal components coordinates
 
@@ -2001,3 +1945,260 @@ class VisualizationFunctions:
         self.saveFigure(fig, self.saveDir, self.dataName + '_energy_landscape_PC%s_vs_PC%s'%(PCy, PCx), extension=extension, dpi=dpi, **kwargs)
 
         return fig
+    
+
+    # Plotly-powered figures
+
+    @tryExcept
+    def makeSankeyDiagram(self, df, colormapForIndex = None, colormapForColumns = None, linksColor = 'rgba(100,100,100,0.6)', title = '', attemptSavingHTML = False, quality = 4, nameAppend = '_Sankey_diagram'):
+
+        '''Make a Sankey diagram, also known as 'river plot' with two groups of nodes
+
+        Parameters:
+            df: pandas.DataFrame 
+                With counts (overlaps)
+
+            colormapForIndex: dictionary, Default None
+                Colors to use for nodes specified in the DataFrame index
+
+            colormapForColumns: dictionary, Default None
+                Colors to use for nodes specified in the DataFrame columns
+
+            linksColor: str, Default 'rgba(100,100,100,0.6)'
+                Color of the non-overlapping links
+
+            title: str, Default ''
+                Title to print on the diagram
+
+            interactive: boolean , Default False
+                Whether to launch interactive JavaScript-based graph
+
+            quality: int, Default 4
+                Proportional to the resolution of the figure to save
+
+            nameAppend: str, Default '_Sankey_diagram'
+                Name to append to the figure file
+
+        Returns:
+            None
+        
+        Usage:
+            DCS = DigitalCellSorter.DigitalCellSorter()
+
+            DCS.makeSankeyDiagram(df)
+        '''
+
+        try:
+            temp_index = pd.MultiIndex.from_arrays([df.index, [colormapForIndex[item] for item in df.index]], names=['label', 'color'])
+            temp_columns = pd.MultiIndex.from_arrays([df.columns, [colormapForColumns[item] for item in df.columns]], names=['label', 'color'])
+            df.index = temp_index
+            df.columns = temp_columns
+        except Exception as exception:
+            print(exception)
+            print('Using default node colors')
+            colormapForIndex = None
+            colormapForColumns = None
+
+        if (colormapForIndex is None) or (colormapForColumns is None):
+            nodeColors = ['rgba(150,0,10,0.8)'] * len(df.index) + ['rgba(10,0,150,0.8)'] * len(df.columns)
+            nodeLabels = df.index.to_list() + df.columns.to_list()
+        else:
+            nodeLabels = df.index.get_level_values('label').to_list() + df.columns.get_level_values('label').to_list()
+            nodeColors = df.index.get_level_values('color').to_list() + df.columns.get_level_values('color').to_list()
+
+        sources, targets, values, labels = [], [], [], []
+        for i, item in enumerate(df.index):
+            sources.extend([i] * len(df.loc[item]))
+            targets.extend(list(range(len(df.index), len(df.index) + len(df.loc[item]))))
+            values.extend([j for j in df.loc[item].values])
+            #labels.extend([item + ' -> ' + jtem for jtem in df.loc[item].index])
+
+        colorscales = [dict(label=label, colorscale=[[0, linksColor], [1, linksColor]]) for label in labels]
+
+        if not nodeColors is None:
+            for i in range(len(sources)):
+                if nodeColors[sources[i]] == nodeColors[targets[i]]:
+                    newColor = ','.join(nodeColors[sources[i]].split(',')[:3] + ['0.6)'])
+                    colorscales[i] = dict(label=labels[i], colorscale=[[0, newColor], [1, newColor]])
+
+        fig = plotly.graph_objects.Figure(data=[plotly.graph_objects.Sankey(valueformat = '', valuesuffix = '',
+            node = dict(pad = 20, thickness = 40, line = dict(color = 'white', width = 0.5), label = nodeLabels, color = nodeColors,),
+            link = dict(source = sources, target = targets, value = values, label = labels, colorscales = colorscales, hoverinfo='all'))]) #line ={'color':'rgba(255,0,0,0.8)', 'width':0.1}
+
+        if not title is None:
+            fig.update_layout(title_text=title, font_size=10)
+
+        try:
+            fig.write_image(os.path.join(self.saveDir, self.dataName + nameAppend + '.png'), width=500, height=500, scale=quality)
+
+        except Exception as exception:
+            print('Cannot save static image (likely due to missing orca). Saving to interactive html')
+            attemptSavingHTML = True
+
+        if attemptSavingHTML:
+            plot_offline(fig, filename=os.path.join(self.saveDir, self.dataName + nameAppend + '.html'), auto_open=False)
+
+        return fig
+
+    @tryExcept
+    def HopfieldLandscapePlot3D(self, PCx = 0, PCy = 1, colorbar = True, fontsize = 12, plotMesh = True, plotAttractors = True, trPath = None, attemptSavingHTML=False, nameAppend = '', quality = 4, **kwargs):
+
+        '''Make heatmap plot of the attractors in their two principal components coordinates
+
+        Parameters:
+            legend: boolean, Default False
+                Whether to add legend containing cell types names
+            
+            labels: boolean, Default False
+                Whether to add labels 
+                
+            PCx: int, Default 0
+                Principal component for x-coordinate of the plot
+                            
+            PCy: int, Default 1
+                Principal component for y-coordinate of the plot
+            
+            colorbar: boolean, Default False
+                Whether to add colorbar
+
+            fontsize: int, Default 10
+                Text labels font size
+            
+            plotMesh: boolean, Default False
+                Whether to plot landscape heatmap
+                
+            plotAttractors: boolean, Default False
+                Whether to plot attractor stars
+                
+            adjustText: boolean, Default False
+                Whether to minimize text labels overlap
+                
+            axisOff: boolean, Default False
+                Whether to hide the axes lines
+
+            colorbarva: float, Default 0.75
+                Vertical position of the bottom of the colorbar
+
+            colorbarha: float, Default 0.85
+                Horizontal position of the colorbar
+
+            trPath: str, Default None
+                Path to trajectories files
+
+            colormap: matplotlib.colormap or str, Default matplotlib.colors.LinearSegmentedColormap.from_list('cmap', [(1, 1, 1), (0, 1, 1), (0, 0, 1), (1, 0, 0)], N=1000)
+                Colormap or its string name
+
+            dpi: int, Default 300
+                Resolution of the figure
+
+            extension: str, Default 'png'
+                Format extension of the figure
+
+        Returns:
+            None
+        
+        Usage:
+            DCS = DigitalCellSorter.DigitalCellSorter()
+
+            DCS.makeHopfieldLandscapePlot()
+        '''
+           
+        if trPath is None:
+            trPath = os.path.join(self.saveDir, 'HopfieldTrajectories')
+
+        if not os.path.exists(trPath):
+            print('Data not found', flush=True)
+
+            return
+
+        attrs_xpca, attrs_names = read(os.path.join(trPath, 'attrs'))
+        attrs_xpca = attrs_xpca[:attrs_xpca.shape[1]]
+
+        mesh_xpca = read(os.path.join(trPath, 'mesh'))
+
+        mesh_energy = mesh_xpca[range(len(mesh_xpca)), -1]
+        mesh_xpca = mesh_xpca[range(len(mesh_xpca)), :-1]
+
+        data = np.vstack([attrs_xpca, mesh_xpca])
+
+        coords = data.T[[PCx, PCy], :]
+        attrs2D, mesh2D = coords[:, :attrs_xpca.shape[1]].T, coords[:, attrs_xpca.shape[1]:].T
+    
+        vmin, vmax = min(mesh_energy), 0.
+
+        vals = mesh_energy.copy()
+        vals[np.where(vals > (vmax - 0.001))[0]] = vmax - 0.001
+        vals[np.where(vals < (vmin + 0.001))[0]] = vmin + 0.001
+
+        xmin, xmax = mesh2D.T[0].min(), mesh2D.T[0].max()
+        ymin, ymax = mesh2D.T[1].min(), mesh2D.T[1].max()
+
+        dx = (xmax - xmin) * 0.05
+        dy = (ymax - ymin) * 0.05
+
+        xmin -= dx
+        xmax += dx
+        ymin -= dy
+        ymax += dy
+
+        ngrid = 100
+        grid = np.zeros((ngrid + 1, ngrid + 1))
+        grid[:] = vmin
+
+        i = (ngrid * (mesh2D.T[0] - xmin) / (xmax - xmin)).astype(int)
+        j = (ngrid * (mesh2D.T[1] - ymin) / (ymax - ymin)).astype(int)
+
+        se = pd.Series(index=zip(i,j), data=mesh_energy).groupby(axis=0, level=0).agg(np.min)
+        se.index = pd.MultiIndex.from_tuples(se.index)
+
+        grid[(se.index.get_level_values(0).values, se.index.get_level_values(1).values)] = se.values
+
+        df = se.unstack(fill_value=vmin)
+
+        fig = go.Figure()
+
+        if plotMesh:
+            fig.add_trace(go.Surface(x=np.linspace(xmin, xmax, df.shape[0]), 
+                                        y=np.linspace(ymin, ymax, df.shape[1]), 
+                                        z=gaussian_filter(df.values.T, sigma=0.75), opacity=1., colorscale="blackbody_r",
+                                        showscale=colorbar,
+                                        hoverinfo='none',
+                                        contours= {'x': {'highlight': False}, 
+                                                'y': {'highlight': False}, 
+                                                'z': {'highlight': False}},))
+
+            fig.update_traces(contours_z=dict(show=True, width=3., highlightwidth=3., usecolormap=False, highlightcolor="limegreen", project=dict(x=True,y=True,z=True), highlight=True, color='grey', size=(vmax-vmin)/10.))
+
+        annotations = []
+        if plotAttractors:
+            for i, point in enumerate(zip(attrs2D.T[0], attrs2D.T[1])):
+                fig.add_trace(go.Scatter3d(x=[point[0], point[0]], y=[point[1], point[1]], z=[vmin, 0.5*vmin],  mode='lines', hoverinfo='none', line=dict(width=2, color='blue'), showlegend=False))
+
+                annotations.append(dict(showarrow=False, x=point[0], y=point[1], z=0.4*vmin, text=attrs_names[i], xanchor="center", xshift=10, opacity=1, font=dict(color='black', size=fontsize)))
+
+            fig.add_trace(go.Scatter3d(x=attrs2D.T[0], y=attrs2D.T[1], z=0. * attrs2D.T[0] + 0.5*vmin, mode='markers', 
+                                        hovertext=attrs_names,
+                                        hoverinfo='text',
+                                        marker=dict(size=5, color='blue'),
+                                        projection=dict(z=dict(show=True)),
+                                        showlegend=False))
+
+        fig.update_layout(title='Hopfield Attractors', autosize=False, width=700, height=700, margin=dict(l=75, r=75, b=75, t=90))
+
+        fig.update_layout(scene = {'xaxis': {'title_text': 'PC1', 'nticks': 10, 'spikesides': False, 'showspikes': False, 'showbackground': False, 'showline': False, 'showticklabels': False, 'showaxeslabels': False}, 'yaxis': {'title_text': 'PC2', 'nticks': 10, 'spikesides': False, 'showspikes': False, 'showbackground': False, 'showline': False, 'showticklabels': False, 'showaxeslabels': False}, 'zaxis': {'title_text': 'Energy', 'range': (vmin, 0.), 'nticks': 10, 'showspikes': False, 'showbackground': False, 'showline': False, 'showticklabels': False, 'showaxeslabels': False}, 'aspectratio': {'x': 1, 'y': 1, 'z': 0.33}, 'annotations': annotations})
+
+        fig.update_layout(scene_camera=dict(up=dict(x=0, y=0, z=2), center=dict(x=0, y=0, z=0), eye=dict(x=0, y=-0.25, z=1.25)))
+
+        fileName = self.dataName + '_energy_landscape_PC%s_vs_PC%s'%(PCy, PCx) + nameAppend
+
+        try:
+            fig.write_image(os.path.join(self.saveDir, fileName + '.png'), width=700, height=700, scale=quality)
+
+        except Exception as exception:
+            print('Cannot save static image (likely due to missing orca). Saving to interactive html')
+            attemptSavingHTML = True
+
+        if attemptSavingHTML:
+            plot_offline(fig, filename=os.path.join(self.saveDir, fileName + '.html'), auto_open=False)
+
+        return fig
